from __future__ import annotations

from soda.contracts.contract import Contract, ContractResult
from soda.contracts.impl.warehouse import Warehouse
from soda.contracts.impl.yaml_helper import YamlFile


class VerificationWarehouse:
    """
    Groups all contracts for a specific warehouse. Used during contract verification execution to group all
    contracts per warehouse and ensure the warehouse is open during verification of the contract for this warehouse.
    """

    def __init__(self) -> None:
        self.warehouse: Warehouse | None = None
        self.contracts: list[Contract] = []

    def requires_with_block(self) -> bool:
        return True

    def add_contract(self, contract: Contract) -> None:
        self.contracts.append(contract)

    def ensure_open_and_verify_contracts(self) -> list[ContractResult]:
        """
        Ensures that the data source has an open connection and then invokes self.__verify_contracts()
        """
        if self.requires_with_block():
            with self.warehouse as d:
                return self.verify_contracts()
        else:
            return self.verify_contracts()

    def verify_contracts(self):
        """
        Assumes the data source has an open connection
        """
        contract_results: list[ContractResult] = []
        for contract in self.contracts:
            contract_result: ContractResult = contract.verify()
            contract_results.append(contract_result)
        return contract_results


class FileVerificationWarehouse(VerificationWarehouse):
    def __init__(self, warehouse_yaml_file: YamlFile):
        super().__init__()
        self.warehouse_file: YamlFile = warehouse_yaml_file
        self.warehouse = Warehouse.from_yaml_file(self.warehouse_file)


class SparkVerificationWarehouse(VerificationWarehouse):
    def __init__(self, spark_session: object, warehouse_name: str = "spark_ds"):
        super().__init__()
        self.spark_session: object = spark_session
        self.warehouse_name = warehouse_name
        self.warehouse = Warehouse.from_spark_session(spark_session=self.spark_session)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod

import soda.common.logs as soda_common_logs
from soda.execution.data_source import DataSource

from soda.contracts.impl.logs import Logs
from soda.contracts.impl.yaml_helper import YamlFile, YamlHelper

logger = logging.getLogger(__name__)


class Warehouse:
    """
    Represents the configurations to create a connection. Usually it's loaded from a YAML file.
    """

    def __init__(self, logs: Logs | None = None):
        self.logs: Logs = logs if logs else Logs()

        # only initialized after the .open() method is called
        self.dbapi_connection: object | None = None
        # only initialized after the .open() method is called
        self.warehouse_name: str | None = None
        # only initialized after the .open() method is called
        self.warehouse_type: str | None = None

    @classmethod
    def from_yaml_file(cls, warehouse_file: YamlFile) -> Warehouse:
        return FileClWarehouse(warehouse_yaml_file=warehouse_file)

    @classmethod
    def from_spark_session(cls, spark_session, logs: Logs | None = None) -> Warehouse:
        return SparkSessionClWarehouse(spark_session=spark_session, logs=logs)

    def __enter__(self) -> Warehouse:
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except Exception as e:
            logger.warning(f"Could not close connection: {e}")

    def __str__(self) -> str:
        return self.warehouse_name

    def open(self) -> None:
        self.dbapi_connection = self._create_dbapi_connection()

    @abstractmethod
    def _create_dbapi_connection(self) -> object:
        pass

    def close(self) -> None:
        """
        Closes te connection. This method will not throw any exceptions.
        Check errors with has_errors or assert_no_errors.
        """
        if self.dbapi_connection:
            try:
                self.dbapi_connection.close()
            except Exception as e:
                logger.warning(f"Could not close the dbapi connection: {e}")


class ClWarehouse(Warehouse, ABC):

    def __init__(self, logs: Logs):
        super().__init__(logs)
        self.sodacl_data_source: DataSource | None = None

    def _create_dbapi_connection(self) -> object:
        self.sodacl_data_source: DataSource = self._create_sodacl_data_source()
        try:
            self.sodacl_data_source.connect()
        except Exception as e:
            self.logs.error(f"Could not connect to '{self.warehouse_name}': {e}")
        return self.sodacl_data_source.connection

    @abstractmethod
    def _create_sodacl_data_source(self) -> DataSource:
        pass


class FileClWarehouse(ClWarehouse):

    def __init__(self, warehouse_yaml_file: YamlFile):
        super().__init__(warehouse_yaml_file.logs)
        self.warehouse_file: YamlFile = warehouse_yaml_file
        self.connection_dict: dict | None = None

        if self.warehouse_file.is_ok():
            yaml_helper: yaml_helper = YamlHelper(yaml_file=self.warehouse_file, logs=self.logs)
            warehouse_yaml_dict: dict = self.warehouse_file.dict
            self.warehouse_type = yaml_helper.read_string(warehouse_yaml_dict, "type")
            self.warehouse_name = yaml_helper.read_string(warehouse_yaml_dict, "name")
            self.connection_dict: dict = yaml_helper.read_dict(warehouse_yaml_dict, "connection")

    def _create_sodacl_data_source(self) -> DataSource:
        # consider translating postgres schema search_path option
        # options = f"-c search_path={schema}" if schema else None
        try:
            return DataSource.create(
                logs=soda_common_logs.Logs(logger=logger),
                data_source_name=self.warehouse_name,
                data_source_type=self.warehouse_type,
                data_source_properties=self.connection_dict,
            )
        except Exception as e:
            self.logs.error(message=f"Could not create the data source: {e}", exception=e)


class SparkSessionClWarehouse(ClWarehouse):

    def __init__(self, spark_session: object, logs: Logs):
        super().__init__(logs)
        self.spark_session: object = spark_session
        self.warehouse_name = "spark_ds"
        self.warehouse_type = "spark_df"

    def _create_sodacl_data_source(self) -> DataSource:
        try:
            return DataSource.create(
                logs=soda_common_logs.Logs(logger=logger),
                data_source_name=self.warehouse_name,
                data_source_type=self.warehouse_type,
                data_source_properties={"spark_session": self.spark_session},
            )
        except Exception as e:
            self.logs.error(message=f"Could not create the spark session data source: {e}", exception=e)



from __future__ import annotations

from typing import Iterator

from soda.contracts.contract import Contract, ContractResult
from soda.contracts.impl.contract_verification_impl import (
    FileVerificationWarehouse,
    SparkVerificationWarehouse,
    VerificationWarehouse,
)
from soda.contracts.impl.logs import Logs
from soda.contracts.impl.soda_cloud import SodaCloud
from soda.contracts.impl.yaml_helper import YamlFile


class ContractVerificationBuilder:

    def __init__(self):
        self.logs: Logs = Logs()
        self.warehouse_yaml_files: list[YamlFile] = []
        self.warehouse_spark_sessions: dict[str, object] = {}
        self.contract_files: list[YamlFile] = []
        self.soda_cloud_files: list[YamlFile] = []
        self.variables: dict[str, str] = {}

    def with_contract_yaml_file(self, contract_yaml_file_path: str) -> ContractVerificationBuilder:
        if not isinstance(contract_yaml_file_path, str):
            self.logs.error(
                message=f"In ContractVerificationBuilder, parameter contract_yaml_file_path must be a string, but was {contract_yaml_file_path} ({type(contract_yaml_file_path)})"
            )
        self.contract_files.append(YamlFile(yaml_file_path=contract_yaml_file_path, logs=self.logs))
        return self

    def with_contract_yaml_str(self, contract_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(contract_yaml_str, str)
        self.contract_files.append(YamlFile(yaml_str=contract_yaml_str, logs=self.logs))
        return self

    def with_contract_yaml_dict(self, contract_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(contract_yaml_dict, dict)
        self.contract_files.append(YamlFile(yaml_dict=contract_yaml_dict, logs=self.logs))
        return self

    def with_warehouse_yaml_file(self, warehouses_yaml_file_path: str) -> ContractVerificationBuilder:
        assert isinstance(warehouses_yaml_file_path, str)
        warehouse_yaml_file = YamlFile(yaml_file_path=warehouses_yaml_file_path, logs=self.logs)
        self.warehouse_yaml_files.append(warehouse_yaml_file)
        return self

    def with_warehouse_yaml_str(self, warehouses_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(warehouses_yaml_str, str)
        warehouse_yaml_file = YamlFile(logs=self.logs, yaml_str=warehouses_yaml_str)
        self.warehouse_yaml_files.append(warehouse_yaml_file)
        return self

    def with_warehouse_yaml_dict(self, warehouses_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(warehouses_yaml_dict, dict)
        warehouse_yaml_file = YamlFile(logs=self.logs, yaml_dict=warehouses_yaml_dict)
        self.warehouse_yaml_files.append(warehouse_yaml_file)
        return self

    def with_warehouse_spark_session(
        self, spark_session: object, warehouse_name: str = "spark_ds"
    ) -> ContractVerificationBuilder:
        assert isinstance(spark_session, object) and isinstance(warehouse_name, str)
        self.warehouse_spark_sessions[warehouse_name] = spark_session
        return self

    def with_variable(self, key: str, value: str) -> ContractVerificationBuilder:
        self.variables[key] = value
        return self

    def with_variables(self, variables: dict[str, str]) -> ContractVerificationBuilder:
        if isinstance(variables, dict):
            self.variables.update(variables)
        return self

    def with_soda_cloud_yaml_file(self, soda_cloud_yaml_file_path: str) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_file_path, str)
        self.soda_cloud_files.append(YamlFile(yaml_file_path=soda_cloud_yaml_file_path, logs=self.logs))
        return self

    def with_soda_cloud_yaml_str(self, soda_cloud_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_str, str)
        self.soda_cloud_files.append(YamlFile(yaml_str=soda_cloud_yaml_str, logs=self.logs))
        return self

    def with_soda_cloud_yaml_dict(self, soda_cloud_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_dict, dict)
        self.soda_cloud_files.append(YamlFile(yaml_dict=soda_cloud_yaml_dict, logs=self.logs))
        return self

    def build(self) -> ContractVerification:
        return ContractVerification(contract_verification_builder=self)

    def execute(self) -> ContractVerificationResult:
        contract_verification: ContractVerification = self.build()
        return contract_verification.execute()


class VerificationWarehouses:
    def __init__(self, contract_verification_builder: ContractVerificationBuilder):
        self.verification_warehouses_by_name: dict[str, VerificationWarehouse] = {}
        # The purpose of the undefined verification warehouse is to ensure that we still capture
        # all the parsing errors in the logs, even if there is no warehouse associated
        self.undefined_verification_warehouse = VerificationWarehouse()
        self.single_verification_warehouse: VerificationWarehouse | None = None
        self.logs: Logs = contract_verification_builder.logs

        # Parse data sources
        variables: dict[str, str] = contract_verification_builder.variables
        for warehouse_yaml_file in contract_verification_builder.warehouse_yaml_files:
            warehouse_yaml_file.parse(variables)
            verification_warehouse: VerificationWarehouse = FileVerificationWarehouse(
                warehouse_yaml_file=warehouse_yaml_file
            )
            self.add(verification_warehouse)
        for warehouse_name, spark_session in contract_verification_builder.warehouse_spark_sessions.items():
            verification_warehouse: VerificationWarehouse = SparkVerificationWarehouse(
                spark_session=spark_session, warehouse_name=warehouse_name
            )
            self.add(verification_warehouse)

    def get(self, contract_warehouse_name: str | None) -> VerificationWarehouse:
        if isinstance(contract_warehouse_name, str):
            verification_warehouse = self.verification_warehouses_by_name.get(contract_warehouse_name)
            if verification_warehouse:
                return verification_warehouse
            else:
                self.logs.error(f"Data source '{contract_warehouse_name}' not configured")
                return self.undefined_verification_warehouse

        if self.single_verification_warehouse:
            # no data source specified in the contract
            # and a single data source was specified in the verification
            return self.single_verification_warehouse

        return self.undefined_verification_warehouse

    def add(self, verification_warehouse: VerificationWarehouse):
        self.verification_warehouses_by_name[verification_warehouse.warehouse.warehouse_name] = verification_warehouse
        # update self.single_verification_warehouse because the number of verification warehouses changed
        warehouses_count = len(self.verification_warehouses_by_name)
        if warehouses_count == 1:
            self.single_verification_warehouse = next(iter(self.verification_warehouses_by_name.values()))
        elif warehouses_count == 0:
            self.single_verification_warehouse = self.undefined_verification_warehouse

    def __iter__(self) -> Iterator[VerificationWarehouse]:
        return iter(self.verification_warehouses_by_name.values())


class ContractVerification:

    @classmethod
    def builder(cls) -> ContractVerificationBuilder:
        return ContractVerificationBuilder()

    def __init__(self, contract_verification_builder: ContractVerificationBuilder):
        self.logs: Logs = contract_verification_builder.logs
        self.variables: dict[str, str] = contract_verification_builder.variables
        self.contracts: list[Contract] = []
        self.contract_results: list[ContractResult] = []
        self.soda_cloud: SodaCloud | None = None
        self.verification_warehouses = self._parse_verification_warehouses(contract_verification_builder)

        # parse the contract files and add them to the matching verification data source
        for contract_file in contract_verification_builder.contract_files:
            contract_file.parse(self.variables)
            if contract_file.is_ok():
                contract_warehouse_name: str | None = contract_file.dict.get("warehouse")

                verification_warehouse = self.verification_warehouses.get(contract_warehouse_name)

                contract: Contract = Contract.create(
                    warehouse=verification_warehouse.warehouse,
                    contract_file=contract_file,
                    variables=self.variables,
                    soda_cloud=self.soda_cloud,
                    logs=contract_file.logs,
                )
                verification_warehouse.add_contract(contract)
                self.contracts.append(contract)

        for soda_cloud_file in contract_verification_builder.soda_cloud_files:
            if soda_cloud_file.exists():
                soda_cloud_file.parse(self.variables)
                self.soda_cloud = SodaCloud(soda_cloud_file)
                break

    def _parse_verification_warehouses(self, contract_verification_builder) -> VerificationWarehouses:
        return VerificationWarehouses(contract_verification_builder)

    def __str__(self) -> str:
        return str(self.logs)

    def execute(self) -> ContractVerificationResult:
        all_contract_results: list[ContractResult] = []
        for verification_warehouse in self.verification_warehouses:
            warehouse_contract_results: list[ContractResult] = verification_warehouse.ensure_open_and_verify_contracts()
            all_contract_results.extend(warehouse_contract_results)
        return ContractVerificationResult(
            logs=self.logs, variables=self.variables, contract_results=all_contract_results
        )


class ContractVerificationResult:
    def __init__(self, logs: Logs, variables: dict[str, str], contract_results: list[ContractResult]):
        self.logs: Logs = logs
        self.variables: dict[str, str] = variables
        self.contract_results: list[ContractResult] = contract_results

    def failed(self) -> bool:
        """
        Returns True if there are execution errors or if there are check failures.
        """
        return not self.passed()

    def passed(self) -> bool:
        """
        Returns True if there are no execution errors and no check failures.
        """
        return not self.logs.has_errors() and all(contract_result.passed() for contract_result in self.contract_results)

    def has_errors(self) -> bool:
        return self.logs.has_errors()

    def has_failures(self) -> bool:
        return any(contract_result.failed() for contract_result in self.contract_results)

    def is_ok(self) -> bool:
        return not self.has_errors() and not self.has_failures()

    def assert_ok(self) -> ContractVerificationResult:
        errors_str: str | None = self.logs.get_errors_str() if self.logs.get_errors() else None
        if errors_str or any(contract_result.failed() for contract_result in self.contract_results):
            raise SodaException(message=errors_str, contract_verification_result=self)
        return self

    def __str__(self) -> str:
        blocks: list[str] = [str(self.logs)]
        for contract_result in self.contract_results:
            blocks.extend(self.__format_contract_results_with_heading(contract_result))
        return "\n".join(blocks)

    @classmethod
    def __format_contract_results_with_heading(cls, contract_result: ContractResult) -> list[str]:
        return [f"# Contract results for {contract_result.contract.dataset}", str(contract_result)]


class SodaException(Exception):
    """
    See also adr/03_exceptions_vs_error_logs.md
    """

    def __init__(
        self, message: str | None = None, contract_verification_result: ContractVerificationResult | None = None
    ):
        self.contract_verification_result: ContractVerificationResult | None = contract_verification_result
        message_parts: list[str] = []
        if message:
            message_parts.append(message)
        if self.contract_verification_result:
            message_parts.append(str(self.contract_verification_result))
        exception_message: str = "\n".join(message_parts)
        super().__init__(exception_message)
